package ru.dbotthepony.kstarbound.server

import io.netty.bootstrap.ServerBootstrap
import io.netty.channel.Channel
import io.netty.channel.ChannelFuture
import io.netty.channel.ChannelInitializer
import io.netty.channel.local.LocalAddress
import io.netty.channel.local.LocalServerChannel
import io.netty.channel.socket.nio.NioServerSocketChannel
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.network.ConnectionType
import java.io.Closeable
import java.net.SocketAddress
import java.util.*
import java.util.concurrent.locks.ReentrantLock
import kotlin.concurrent.withLock

class ServerChannels(val server: StarboundServer) : Closeable {
	private val channels = Collections.synchronizedList(ArrayList<ChannelFuture>())
	private val connections = Collections.synchronizedList(ArrayList<ServerConnection>())
	private var localChannel: Channel? = null
	private val lock = ReentrantLock()
	private var isClosed = false

	@Suppress("name_shadowing")
	fun createLocalChannel(): Channel {
		val localChannel = localChannel

		if (localChannel != null) {
			return localChannel
		}

		lock.withLock {
			val localChannel = this.localChannel

			if (localChannel != null) {
				return localChannel
			}

			val channel = ServerBootstrap().channel(LocalServerChannel::class.java).group(Connection.NIO_POOL).childHandler(object : ChannelInitializer<Channel>() {
				override fun initChannel(ch: Channel) {
					LOGGER.info("Incoming connection from ${ch.remoteAddress()}")
					val connection = ServerConnection(server, ConnectionType.MEMORY)
					connections.add(connection)
					connection.bind(ch)
				}
			}).bind(LocalAddress.ANY).syncUninterruptibly()

			channels.add(channel)
			this.localChannel = channel.channel()

			channel.channel().closeFuture().addListener {
				channels.remove(channel)
				this.localChannel = null
			}

			return channel.channel()
		}
	}

	fun createChannel(localAddress: SocketAddress): Channel {
		lock.withLock {
			val channel = ServerBootstrap().channel(NioServerSocketChannel::class.java).group(Connection.NIO_POOL).childHandler(object : ChannelInitializer<Channel>() {
				override fun initChannel(ch: Channel) {
					LOGGER.info("Incoming connection from ${ch.remoteAddress()}")
					val connection = ServerConnection(server, ConnectionType.NETWORK)
					connections.add(connection)
					connection.bind(ch)
				}
			}).bind(localAddress).syncUninterruptibly()

			channels.add(channel)
			this.localChannel = channel.channel()

			channel.channel().closeFuture().addListener {
				channels.remove(channel)
			}

			return channel.channel()
		}
	}

	override fun close() {
		lock.withLock {
			if (isClosed) return

			connections.forEach { it.disconnect("Server is stopping") }
			channels.forEach { it.channel().close() }
			channels.clear()
			connections.clear()
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
