package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World

/**
 * (Hopefully) Static world entities (Plants, Objects, etc), which reside on cell grid
 */
abstract class TileEntity(path: String) : AbstractEntity(path) {
	private var forceChunkRepos = false

	var position = Vector2i()
		set(value) {
			val old = field

			if (isSpawned) {
				field = world.geometry.wrap(value)

				val oldChunkPos = world.geometry.chunkFromCell(old)
				val newChunkPos = world.geometry.chunkFromCell(field)

				chunkPos = newChunkPos

				if (oldChunkPos != newChunkPos || forceChunkRepos) {
					chunk = world.chunkMap[newChunkPos]
					forceChunkRepos = false
				}
			} else {
				field = value
			}
		}

	final override var chunkPos: ChunkPos = ChunkPos.ZERO
		private set

	override fun onSpawn(world: World<*, *>) {
		world.tileEntities.add(this)
		forceChunkRepos = true
		position = position
	}

	override fun onRemove(world: World<*, *>) {
		world.tileEntities.remove(this)
	}
}
